import React, {useCallback, useContext, useEffect, useMemo, useRef, useState} from "react";
import ReactDOM from "react-dom";
import {v4 as uuidv4} from "uuid";
import {CurtainInstance} from "./CurtainInstance";

/**
 * Curtain UUID type.
 */
export type CurtainUuidType = string;

/**
 * The function that opens a curtain.
 */
export type OpenCurtainFunction = (content: React.ReactNode) => CurtainUuidType;

/**
 * The function that closes a curtain.
 */
export type CloseCurtainFunction = (uuid: CurtainUuidType) => void;

/**
 * The function that checks if a curtain is closed (while transitioning out) or not.
 */
export type IsCurtainClosedFunction = (uuid: CurtainUuidType) => boolean;

/**
 * Interface of curtains state.
 */
export interface CurtainsContextState
{
	/**
	 * Open a new curtain.
	 * @param content The curtain content.
	 * @return UUID of the curtain.
	 */
	open: OpenCurtainFunction;

	/**
	 * Close the given curtain.
	 * @param uuid UUID of the curtain to close.
	 */
	close: CloseCurtainFunction;

	/**
	 * Check if the given curtain is closed or not.
	 * @param uuid UUID of the curtain to check.
	 */
	isClosed: IsCurtainClosedFunction;
}

const CurtainsContext = React.createContext<CurtainsContextState>({
	// Empty functions.
	open() { return ""; },
	close() {},
	isClosed() { return false; },
});

/**
 * Hook to interact with curtains (open or close).
 */
export function useCurtains(): CurtainsContextState
{
	return useContext(CurtainsContext);
}



/**
 * A generic callable curtain element.
 */
export type CallableCurtainElement = (...args: any) => React.ReactNode;

/**
 * A callable curtain function with typed parameters.
 */
export type CallableCurtain<F extends CallableCurtainElement> = (...args: Parameters<F>) => CurtainUuidType;

/**
 * Callable curtain function generator.
 * @param curtains The curtains context state.
 * @param curtainElement The curtain element to open when called.
 */
export function callableCurtain<F extends CallableCurtainElement>(curtains: CurtainsContextState, curtainElement: React.ReactNode|F): CallableCurtain<F>
{
	if (typeof curtainElement == "function")
		// It's a callable curtain element, the callable curtain should be called with the same parameters.
		return (...args: Parameters<F>) => curtains.open(curtainElement(...args));
	else
		// It's a simple element, just open it.
		return () => curtains.open(curtainElement);
}

/**
 * Hook to create a simple curtain.
 * @param curtainElement Content of the curtain to open.
 */
export function useCallableCurtain<F extends CallableCurtainElement>(curtainElement: React.ReactNode|F): CallableCurtain<F>
{
	// Get curtains context state.
	const curtains = useCurtains();
	// Create and keep the curtain callable in memory.
	return useCallback(callableCurtain(curtains, curtainElement), [curtains, curtainElement]);
}



/**
 * Page curtains provider.
 */
export function CurtainsProvider({children}: React.PropsWithChildren<{}>)
{
	// Curtains state.
	const [curtains, setCurtains] = useState<Record<CurtainUuidType, React.ReactNode>>({});
	// Keeping track of closed curtains that are still on (while transitioning out).
	const [closedCurtains, setClosedCurtains] = useState<Record<CurtainUuidType, boolean>>({});

	// Initialize open curtain function.
	const open = useRef<OpenCurtainFunction>();
	open.current = useCallback((content: React.ReactNode) => {
		// Generate a new curtain UUID for the new curtain to open.
		const curtainUuid = uuidv4();

		// Add the curtain to open to the list of curtains, with the generated UUID.
		setCurtains({
			...curtains,
			[curtainUuid]: content,
		});

		// Return the curtain UUID.
		return curtainUuid;
	}, [curtains, setCurtains]);

	// Initialize remove curtain function.
	const remove = useRef<CloseCurtainFunction>();
	remove.current = useCallback((uuid: CurtainUuidType) => {
		// Copy the curtains list.
		const newCurtains = {...curtains};
		const newClosedCurtains = {...closedCurtains};
		// Remove the given curtain from the list.
		delete newCurtains[uuid];
		delete newClosedCurtains[uuid];
		// Set the new curtains list.
		setCurtains(newCurtains);
		setClosedCurtains(newClosedCurtains);
	}, [curtains, setCurtains, closedCurtains, setClosedCurtains]);

	// Initialize close curtain function with animation.
	const close = useRef<CloseCurtainFunction>();
	close.current = useCallback((uuid) => {
		// Add the given curtain UUID to the list of closed curtains.
		setClosedCurtains({
			...closedCurtains,
			[uuid]: true,
		});

		// Remove the curtain 1s later.
		window.setTimeout(() => {
			// Remove the curtain.
			remove.current(uuid);
		}, 1000);
	}, [remove, closedCurtains, setClosedCurtains]);

	// Initialize isClosed curtain function.
	const isClosed = useRef<IsCurtainClosedFunction>();
	isClosed.current = useCallback((uuid) => (!!closedCurtains?.[uuid]), [closedCurtains]);

	// Initialize context state from action functions.
	const contextState = useMemo(() => ({
		open: (content: React.ReactNode) => open.current(content),
		close: (uuid: CurtainUuidType) => close.current(uuid),
		isClosed: (uuid: CurtainUuidType) => isClosed.current(uuid),
	}), [open, close, isClosed]);

	// Show dimmed main content.
	useEffect(() => {
		if (
			Object.keys(curtains).filter((curtainId) => !closedCurtains[curtainId]).length > 0
		) // We should dim content if there is at least one open curtain.
		{ // Only dim if it's not already dimmed.
			if (!document.body.classList.contains("dimmed")) document.body.classList.add("dimmed");
		}
		else
			// We shouldn't dim content.
			document.body.classList.remove("dimmed");
	}, [curtains, closedCurtains]);

	return (
		<CurtainsContext.Provider value={contextState}>
			{children}
			<CurtainsPortal curtains={curtains} />
		</CurtainsContext.Provider>
	);
}

/**
 * Curtains portal manager.
 */
function CurtainsPortal({curtains}: {
	curtains: Record<CurtainUuidType, React.ReactNode>;
})
{
	return ReactDOM.createPortal(Object.entries(curtains).map(([uuid, curtainContent]) => (
		<CurtainInstance key={uuid} uuid={uuid}>
			{curtainContent}
		</CurtainInstance>
	)), document.body);
}
